/* Copyright (c) 2003-2009 Dovecot authors, see the included COPYING file */

#include "lib.h"
#include "env-util.h"
#include "array.h"
#include "str.h"
#include "sql-api-private.h"

#ifdef BUILD_ORACLE
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <sqlcpr.h>
#include <sqlca.h>
#include <sqlda.h>
#include <oci.h>

/* Minimum delay between reconnecting to same server */
#define CONNECT_MIN_DELAY 1
/* Maximum time to avoiding reconnecting to same server */
#define CONNECT_MAX_DELAY (60*30)

/* Maximum number of elements in select list */
#define MAX_ITEMS 50
/* Maximum length of then names of the select list items */
#define MAX_VNAME_LEN 50
/* Maximum length of then names of the select list indicators */
#define MAX_INAME_LEN 50

EXEC SQL INCLUDE SQLCA;
EXEC SQL INCLUDE SQLDA;

struct oracle_db {
	struct sql_db api;

	pool_t pool;

	ARRAY_DEFINE(connections, struct oracle_connection);
	unsigned int next_query_connection;
};

struct oracle_connection {
	struct oracle_db *db;

	char *name;
	char *dblogin;

	char *error;

	unsigned int connect_delay;
	unsigned int connect_failure_count;

	time_t last_connect;
	unsigned int connected:1;
};

struct oracle_result {
	struct sql_result api;
	struct oracle_connection *conn;

	char *query;

	unsigned int rownum, rows;

	SQLDA *descriptor;
};

struct oracle_transaction_context {
	struct sql_transaction_context ctx;

	struct oracle_connection *conn;

	pool_t query_pool;
	struct oracle_query_list *head, *tail;

	const char *error;

	unsigned int failed:1;
};

struct oracle_query_list {
	struct oracle_query_list *next;
	const char *query;
	unsigned int *affected_rows;
};

extern struct sql_db driver_oracle_db;
extern struct sql_result driver_oracle_result;
extern struct sql_result driver_oracle_error_result;

static char *from_oracle_strcpy( char *dest,
				int destsize,
				const unsigned char *src,
				unsigned short srclen )
{
	if (destsize!=sizeof(void *)&&destsize<=srclen)
	{
		srclen = destsize - 1;
	}   
	memcpy(dest,src,(size_t)srclen);
	dest[srclen] = '\0';
	return dest;
}

static unsigned char *to_oracle_strcpy( unsigned char *dest,
					unsigned short *destlen,
					int destsize,
					const char *src )
{
	int srclen;
  
	srclen = strlen( src );
	if( srclen>destsize )
	{
		srclen = destsize;
	}     
	*destlen = srclen;
	memcpy( dest, src, srclen );
	return dest;
}

#define TO_ORACLE_STRCPY( dest,src ) \
to_oracle_strcpy( dest.arr,&dest.len,sizeof(dest.arr),src )

#define FROM_ORACLE_STRCPY( dest,src ) \
from_oracle_strcpy( dest,sizeof(dest),src.arr,src.len )

#define SQLERRM_STRCPY( dest ) \
if (sqlca.sqlcode == 0) \
strcpy( dest, "ORA-00000: normal successful completion" ); \
else { \
strncpy( dest, sqlca.sqlerrm.sqlerrmc, sqlca.sqlerrm.sqlerrml ); \
dest[sqlca.sqlerrm.sqlerrml - 1] = '\0'; \
}

extern void SQLColumnNullCheck( dvoid *context, unsigned short *value_type, unsigned short *type_code, int *null_status );
static const char *driver_oracle_escape_string( struct sql_db *_db ATTR_UNUSED, const char *string );

static void driver_oracle_close(struct oracle_connection *conn)
{

	EXEC SQL BEGIN DECLARE SECTION;
		VARCHAR cname[20];
	EXEC SQL END DECLARE SECTION;

	TO_ORACLE_STRCPY(cname, conn->name);

	EXEC SQL AT :cname COMMIT WORK RELEASE;

	conn->connected = FALSE;

	i_free( conn->name );
	i_free( conn->error );
	i_free( conn->dblogin );
}

static bool driver_oracle_connect(struct oracle_connection *conn)
{

	EXEC SQL BEGIN DECLARE SECTION;
		VARCHAR ora_connect_str[256];
		VARCHAR cname[20];
	EXEC SQL END DECLARE SECTION;

	i_info("oraconn: requested connection %s with dblogin %s", conn->name, conn->dblogin );

	TO_ORACLE_STRCPY(cname, conn->name);

	time_t now;

	if (conn->connected) {
		i_info("oraconn: %s is already connected", conn->name);
		return TRUE;
	}

	now = time(NULL);

	if (conn->last_connect + (time_t)conn->connect_delay > now)
		return FALSE;
	conn->last_connect = now;

	TO_ORACLE_STRCPY(ora_connect_str, conn->dblogin);

	EXEC SQL WHENEVER SQLERROR CONTINUE;
	EXEC SQL CONNECT :ora_connect_str AT :cname;

	SQLERRM_STRCPY( conn->error );

	if (sqlca.sqlcode == 0) {
		i_info("oracle: connected.");

		conn->connect_failure_count = 0;
		conn->connect_delay = CONNECT_MIN_DELAY;
		conn->connected = TRUE;
		SQLERRM_STRCPY( conn->error );
		return TRUE;
	} else {
		if (conn->connect_failure_count > 0) {
			/* increase delay between reconnections to this
			   server */
			conn->connect_delay *= 5;
			if (conn->connect_delay > CONNECT_MAX_DELAY)
				conn->connect_delay = CONNECT_MAX_DELAY;
		}
		conn->connect_failure_count++;

		SQLERRM_STRCPY( conn->error );
		i_error("oracle. connect failed: %s", conn->error);
		return FALSE;
	}

}

static int driver_oracle_connect_all(struct sql_db *_db)
{
	struct oracle_db *db = (struct oracle_db *)_db;
	struct oracle_connection *conn;
	unsigned int i, count;
	int ret = -1;

	conn = array_get_modifiable(&db->connections, &count);

	i_info("conall: got %d connections", count);

	for (i = 0; i < count; i++) {
		i_info("conall: connecting number %d", i);
		if (driver_oracle_connect(&conn[i]))
			ret = 1;
	}
	return ret;
}

static char *driver_oracle_generate_name( void )
{
	static unsigned int counter = 0;

	return i_strdup_printf( "ORACONN_%x", ++counter );
}

static void driver_oracle_connection_add( struct oracle_db *db, const char *dblogin )
{
	struct oracle_connection *conn;

	conn = array_append_space(&db->connections);
	conn->db = db;

	i_info( "cadd: connection data initialized");

	conn->name = driver_oracle_generate_name();
	conn->dblogin = p_strdup( db->pool, dblogin );

	i_info( "cadd: created unique name %s for connect string %s", conn->name, conn->dblogin);

	conn->connect_delay = CONNECT_MIN_DELAY;

	i_info( "cadd: set delay %d", conn->connect_delay);

	conn->error = malloc( sizeof(char) * 71 );

	i_info( "cadd: allocated error buffer at %d", *(conn->error));

}

static void driver_oracle_parse_connect_string(struct oracle_db *db,
					      const char *connect_string)
{
	const char *const *args, *name, *value;

	args = t_strsplit_spaces(connect_string, " ");
	for (; *args != NULL; args++) {
		value = strchr(*args, '=');
		if (value == NULL) {
			i_fatal("oracle: Missing value in connect string: %s", *args);
		}
		name = t_strdup_until(*args, value);
		value++;

		if ( strcmp(name, "ora_dblogin") == 0 )
			driver_oracle_connection_add(db, value);
		else if (strcmp(name, "ora_base") == 0)
			env_put( t_strdup_printf( "ORACLE_BASE=%s", value ) );
		else if (strcmp(name, "ora_home") == 0)
			env_put( t_strdup_printf( "ORACLE_HOME=%s", value ) );
		else if (strcmp(name, "ora_tnsadmin") == 0)
			env_put( t_strdup_printf( "TNS_ADMIN=%s", value ) );
		else if (strcmp(name, "ora_sid") == 0)
			env_put( t_strdup_printf( "ORACLE_SID=%s", value ) );
		else if (strcmp(name, "ora_twotask") == 0)
			env_put( t_strdup_printf( "TWO_TASK=%s", value ) );
		else if (strcmp(name, "ora_nlslang") == 0)
			env_put( t_strdup_printf( "NLS_LANG=%s", value ) );
		else
			i_fatal("oracle: Unknown parameter %s in connect string", name);
	}

	if (array_count(&db->connections) == 0)
		i_fatal("oracle: No dblogin given in connect string");
}

static struct sql_db *driver_oracle_init_v(const char *connect_string)
{
	struct oracle_db *db;
	pool_t pool;

	i_info( "initv: oracle initialization: %s", connect_string);

	i_assert(connect_string != NULL);

	srand(time( NULL ));

	pool = pool_alloconly_create("oracle driver", 512);

	db = p_new(pool, struct oracle_db, 1);
	db->pool = pool;
	db->api = driver_oracle_db;
	p_array_init(&db->connections, db->pool, 6);

	T_BEGIN {
		driver_oracle_parse_connect_string(db, connect_string);
	} T_END;

	return &db->api;
}

static void driver_oracle_deinit_v(struct sql_db *_db)
{
	struct oracle_db *db = (struct oracle_db *)_db;
	struct oracle_connection *conn;
	unsigned int i, count;

	conn = array_get_modifiable(&db->connections, &count);
	for (i = 0; i < count; i++)
		(void)driver_oracle_close(&conn[i]);

	array_free(&_db->module_contexts);
	pool_unref(&db->pool);
}

static enum sql_db_flags driver_oracle_get_flags(struct sql_db *db ATTR_UNUSED)
{
	return 0;
}

static const char *driver_oracle_escape_string( struct sql_db *_db ATTR_UNUSED, const char *string )
{
	char *dest, *d;
	const char *p;
	int count = 0;

	/* count "'" entries in source str to malloc correctly */

	for ( p = string; (*p); p++ ) if (*p == '\'') count++;

	dest = malloc( sizeof( char ) * ( strlen( string ) + count + 1 ) );

	p = string;
	d = dest;

	while ( *p ) {
		if ( *p != '\'' ) {
			*d = *p;
		} else {
			*(d++) = '\''; *d = '\'';
		}
		p++; d++;
	}

	*d = '\0';

	return dest;
}

static SQLDA *prepare_descriptor( int max_items, int max_vname_len, int max_iname_len )
{
	SQLDA *dp;
	int i;

	i_info("prepd: requested new SQLDescriptor");
	if ((dp = SQLSQLDAAlloc( SQL_SINGLE_RCTX, max_items, max_vname_len, max_iname_len )) == (SQLDA *)0) {
		i_info("prepd: couln't allocate descriptor");
		return (SQLDA *)0;
        }

        dp->N = max_items;

	for ( i=0; i<dp->N; i++ )
	{
		dp->I[i] = (short *) malloc( sizeof( short ) );
		dp->V[i] = (char *) malloc( 1 );
	}

	i_info("prepd: descriptor created");

	return dp;
}

static struct oracle_connection *get_valid_connection( struct oracle_db *db )
{
	struct oracle_connection *conn;
	unsigned int i, start, count;

	i_info("gvc: requested connection");

	conn = array_get_modifiable( &db->connections, &count );

	i_info("gvc: currently %d conns avail", count);

	start = db->next_query_connection % count;
	db->next_query_connection++;
	i_info("gvc: starting from  %d. Next will be %d", start, db->next_query_connection);

	i=start;
	do {
		i_info("gvc: connecting %d connection", i);
		if ( driver_oracle_connect( &conn[i] ) ) {
			i_info("gvc: connection valid");
			return conn;
		} 
		i = (i+1) % count;
	} while (i!=start);

	i_info("gvc: couldn't find valid connection. resetting delays");

	for ( i=0; i<count; i++ ) conn[i].connect_delay = CONNECT_MIN_DELAY;

	return NULL;
}

static struct sql_result *driver_oracle_exequery_conn( struct sql_db *_db, struct oracle_connection *conn, const char *query)
{
	EXEC SQL BEGIN DECLARE SECTION;
		VARCHAR cname[20];
		VARCHAR sqltext[65535];
	EXEC SQL END DECLARE SECTION;

	struct oracle_result *result;
	int i, precision, scale, null_ok;

	SQLDA *select_dp;

	result = i_new( struct oracle_result, 1 );
	result->api = driver_oracle_result;
	result->api.db = _db;
	result->conn = conn;

	if (conn == NULL) {
		i_info("exeqry: no connection specified");
		strcpy( conn->error, "No connection while execute query" );
		result->api = driver_oracle_error_result;
                return &result->api;
        }

	i_info("exeqry: executing query on conn %s", conn->name);

	if ((select_dp = prepare_descriptor( MAX_ITEMS, MAX_VNAME_LEN, MAX_INAME_LEN )) == (SQLDA *)0) {
		i_info("exeqry: couldn't create SQLDA");
		strcpy( conn->error, "Couln't create SQLDA structure" );
		result->api = driver_oracle_error_result;
		return &result->api;
	}

	result->descriptor = select_dp;

	TO_ORACLE_STRCPY( cname, conn->name );
	TO_ORACLE_STRCPY( sqltext, query );

	i_info("exeqry: all green");

	EXEC SQL WHENEVER SQLERROR CONTINUE;

	EXEC SQL AT :cname DECLARE statement STATEMENT;

	SQLERRM_STRCPY( conn->error );
	if ( sqlca.sqlcode != 0 ) {
	        i_error("oracle. declare statement failed: %s", conn->error);
                result->api = driver_oracle_error_result;
                return &result->api;
	}
	i_info("exeqry: declared statement");


	EXEC SQL AT :cname PREPARE statement FROM :sqltext;

	SQLERRM_STRCPY( conn->error );
	if ( sqlca.sqlcode != 0 ) {
	        i_error("oracle. prepare statement failed: %s", conn->error);
                result->api = driver_oracle_error_result;
                return &result->api;
	}
	i_info("exeqry: statement prepared");

	EXEC SQL AT :cname DECLARE cursor CURSOR FOR statement;

	SQLERRM_STRCPY( conn->error );
	if ( sqlca.sqlcode != 0 ) {
	        i_error("oracle. declare cursor failed: %s", conn->error);
                result->api = driver_oracle_error_result;
                return &result->api;
	}
	i_info("exeqry: declared cursor. about to open");

	EXEC SQL AT :cname OPEN cursor;

	SQLERRM_STRCPY( conn->error );
	if ( sqlca.sqlcode != 0 ) {
	        i_error("oracle. open cursor failed: %s", conn->error);
                result->api = driver_oracle_error_result;
                return &result->api;
	}
	i_info("exeqry: cursor open");

	EXEC SQL AT :cname DESCRIBE SELECT LIST FOR statement INTO select_dp;
	
	SQLERRM_STRCPY( conn->error );
	if ( select_dp->F < 0 ) {
	        i_error("oracle. too many columns: %d while max %d allowed", select_dp->F, MAX_ITEMS);
                result->api = driver_oracle_error_result;
                return &result->api;
	}

	select_dp->N = select_dp->F;
	i_info("exeqry: select list described");

	for (i = 0; i < select_dp->F; i++)
	{
	        SQLColumnNullCheck (SQL_SINGLE_RCTX, (unsigned short *)&(select_dp->T[i]), (unsigned short *)&(select_dp->T[i]), &null_ok);

		*(select_dp->S[i] + select_dp->C[i]) = '\0';

		for (char *p = select_dp->S[i]; (*p = (char)tolower(*p)); p++);

		i_info("exeqry: field [%s] - %d of %d. Defining length", select_dp->S[i], i, select_dp->F);

		switch (select_dp->T[i])
		{
			case  1 :	/* CHAR */
					i_info("exeqry: type char. length %d", select_dp->L[i]);
					break;
			case  2 :	/* NUMBER */
					SQLNumberPrecV6( SQL_SINGLE_RCTX, 
					(unsigned int *)&(select_dp->L[i]), &precision, &scale);
					if (precision == 0) precision = 40;
					if (scale > 0)
					    select_dp->L[i] = sizeof(float);
					else
					    select_dp->L[i] = sizeof(int);
					i_info("exeqry: type number (%d,%d). length %d", precision, scale, select_dp->L[i]);
					break;

			case  8 : /* LONG datatype */
					select_dp->L[i] = 240;
					i_info("exeqry: type long. length %d", select_dp->L[i]);
					break;

			case 11 : /* ROWID datatype */
					select_dp->L[i] = 18;
					i_info("exeqry: type rowid. length %d", select_dp->L[i]);
					break;

			case 12 : /* DATE datatype */
					select_dp->L[i] = 9;
					i_info("exeqry: type date. length %d", select_dp->L[i]);
					break;
 
			case 23 : /* RAW datatype */
					i_info("exeqry: type raw. length %d", select_dp->L[i]);
					break;

			case 24 : /* LONG RAW datatype */
					select_dp->L[i] = 240;
					i_info("exeqry: type longraw. length %d", select_dp->L[i]);
					break;
		}
		if (select_dp->T[i] != 2)
			select_dp->V[i] = (char *) realloc(select_dp->V[i], select_dp->L[i] + 1);  
		else
			select_dp->V[i] = (char *) realloc(select_dp->V[i], select_dp->L[i]);  

		i_info("exeqry: reallocated");

		/* Coerce all datatypes to STRING including LONG RAW and NUMBER */
		select_dp->T[i] = 5;

		i_info("exeqry: coerced all non-numeric datatypes to char");
	}


	i_info("exeqry: set descriptor and connection in result. all done");
	return &result->api;

}

static struct sql_result *driver_oracle_query_s( struct sql_db *_db, const char *query )
{
	i_info("query_s execution requested: %s", query);

	return driver_oracle_exequery_conn( _db, get_valid_connection( (struct oracle_db *)_db ), query );
}

static void driver_oracle_query(struct sql_db *db, const char *query,
                               sql_query_callback_t *callback, void *context)
{
        struct sql_result *result;

	i_info("query execution requested: %s", query);

        result = sql_query_s(db, query);
	i_info("query executed. making callback");
        result->callback = TRUE;
        callback(result, context);
        result->callback = FALSE;
	i_info("callback made. freeing result");
        sql_result_free(result);
}

static void driver_oracle_exec( struct sql_db *_db, const char *query )
{
	EXEC SQL BEGIN DECLARE SECTION;
		VARCHAR cname[20];
		VARCHAR sqltext[65535];
	EXEC SQL END DECLARE SECTION;

	struct oracle_db *db = (struct oracle_db *) _db;
	struct oracle_connection *conn;

	i_info("oracle simple execution requested: %s", query);

	conn = get_valid_connection( db );

	TO_ORACLE_STRCPY(cname, conn->name);
	TO_ORACLE_STRCPY(sqltext, query);

	EXEC SQL WHENEVER SQLERROR CONTINUE;

	EXEC SQL AT :cname EXECUTE IMMEDIATE :sqltext;

	SQLERRM_STRCPY( conn->error );
}

static struct sql_transaction_context *driver_oracle_transaction_begin( struct sql_db *db )
{
	struct oracle_transaction_context *ctx;

	i_info("begintran: new transaction requested. creating context");
	ctx = i_new( struct oracle_transaction_context, 1 );
	ctx->ctx.db = db;
	ctx->conn = get_valid_connection( (struct oracle_db *) db );
	ctx->query_pool = pool_alloconly_create( "oracle transaction", 1024 );

	return &ctx->ctx;
}

static void driver_oracle_transaction_rollback( struct sql_transaction_context *_ctx )
{
	EXEC SQL BEGIN DECLARE SECTION;
		VARCHAR cname[20];
	EXEC SQL END DECLARE SECTION;

	struct oracle_transaction_context *ctx = (struct oracle_transaction_context *) _ctx;

	i_info("rbtran: rollback requested");

	TO_ORACLE_STRCPY( cname, ctx->conn->name );

	EXEC SQL WHENEVER SQLERROR CONTINUE;

	EXEC SQL AT :cname ROLLBACK WORK;

	SQLERRM_STRCPY( ctx->conn->error );
	pool_unref( &ctx->query_pool );
	i_free( ctx );
}

static void driver_oracle_update(struct sql_transaction_context *_ctx, const char *query, unsigned int *affected_rows)
{
        struct oracle_transaction_context *ctx = (struct oracle_transaction_context *)_ctx;
        struct oracle_query_list *list;

	i_info("oupdate: query list update. new query: %s", query);

        list = p_new(ctx->query_pool, struct oracle_query_list, 1);
        list->query = p_strdup(ctx->query_pool, query);
        list->affected_rows = affected_rows;

        if (ctx->head == NULL)
                ctx->head = list;
        else
                ctx->tail->next = list;
        ctx->tail = list;
}

static void driver_oracle_transaction_commit(struct sql_transaction_context *ctx, sql_commit_callback_t *callback, void *context)
{
        const char *error;

	i_info("commtran: commit requested");
        if (sql_transaction_commit_s(&ctx, &error) < 0)
                callback(error, context);
        else
                callback(NULL, context);
}

static int transaction_send_query( struct oracle_transaction_context *ctx, const char *query )
{
	struct sql_result *result;
	int ret = 0;

	i_info("sendqry: sending query %s", query);
	if (ctx->failed) return -1;

	result = driver_oracle_exequery_conn( ctx->ctx.db, ctx->conn, query );

	if (sql_result_next_row(result) < 0) {
		ctx->error = sql_result_get_error(result);
		ctx->failed = TRUE;
		ret = -1;
	} else if (ctx->head != NULL && ctx->head->affected_rows != NULL) {
		*ctx->head->affected_rows = sqlca.sqlerrd[2];
	}

	sql_result_free(result);
	return ret;
}

static int driver_oracle_transaction_commit_s(struct sql_transaction_context *_ctx, const char **error_r)
{

	EXEC SQL BEGIN DECLARE SECTION;
		VARCHAR cname[20];
	EXEC SQL END DECLARE SECTION;

        struct oracle_transaction_context *ctx = (struct oracle_transaction_context *)_ctx;

	i_info("comstran: commit requested");

        TO_ORACLE_STRCPY( cname, ctx->conn->name );

	*error_r = NULL;

	while (ctx->head != NULL) {
		if (transaction_send_query(ctx, ctx->head->query) < 0) {
			SQLERRM_STRCPY( ctx->conn->error );
			i_info("comstran: query failed: %s", ctx->conn->error);
			sql_transaction_rollback( &_ctx );
			*error_r = ctx->conn->error;
			return -1;
		}
		ctx->head = ctx->head->next;
	}

	EXEC SQL AT :cname COMMIT WORK;

	SQLERRM_STRCPY( ctx->conn->error );
	*error_r = ctx->conn->error;

	i_info("comstran: commit complete");
        return 0;
}

static void driver_oracle_result_free( struct sql_result *_result )
{

        EXEC SQL BEGIN DECLARE SECTION;
                VARCHAR cname[20];
        EXEC SQL END DECLARE SECTION;

	struct oracle_result *result = (struct oracle_result *) _result;
	int i;

	i_info("resfre: requested to free result area");
	TO_ORACLE_STRCPY( cname, result->conn->name );

	if ( _result->callback ) return;

	if ( result->descriptor != NULL )
	{
		for (i = 0; i < MAX_ITEMS; i++)
		{    
			if (result->descriptor->V[i] != (char *) 0) free(result->descriptor->V[i]);
			free(result->descriptor->I[i]); /* MAX_ITEMS were allocated. */
		}
	}

	SQLSQLDAFree( SQL_SINGLE_RCTX, result->descriptor );

	EXEC SQL WHENEVER SQLERROR CONTINUE;

	EXEC SQL AT :cname CLOSE cursor;
	SQLERRM_STRCPY( result->conn->error );

	i_free( result );
	i_info("resfre: result area free");
}

static int driver_oracle_result_next_row( struct sql_result *_result )
{

        EXEC SQL BEGIN DECLARE SECTION;
                VARCHAR cname[20];
        EXEC SQL END DECLARE SECTION;

	struct oracle_result *result = (struct oracle_result *) _result;
	SQLDA *select_dp = result->descriptor;
	int i;

	i_info("nextrow: next row requested for connection %s", result->conn->name);

        TO_ORACLE_STRCPY( cname, result->conn->name );

	if (select_dp == NULL) return 0;

	i_info("nextrow: select list seems to be valid and involves %d fields", select_dp->F);

	EXEC SQL WHENEVER SQLERROR GOTO fail;

	EXEC SQL WHENEVER NOT FOUND GOTO notfound;

	i_info("nextrow: about to fetch");
	EXEC SQL AT :cname FETCH cursor USING DESCRIPTOR select_dp;

	i_info("nextrow: fetch successful");
	SQLERRM_STRCPY( result->conn->error );

	for (i = 0; i < select_dp->F; i++) {
		i_info("nextrow: field %d [%s] - (%d) [%s]", i, select_dp->S[i], select_dp->L[i], select_dp->V[i]);
	}

	return 1;

notfound:	i_info("nextrow: notfound while fetching. cursor exhausted");
		SQLERRM_STRCPY( result->conn->error );
		return 0;

fail:		SQLERRM_STRCPY( result->conn->error );
		i_info("nextrow: %s", result->conn->error);
		return -1;

}

static unsigned int driver_oracle_result_get_fields_count( struct sql_result *_result )
{
	struct oracle_result *result = (struct oracle_result *) _result;
	i_info("gfc: fields count requested");

	return (unsigned int)result->descriptor->F;
}

static const char *driver_oracle_result_get_field_name( struct sql_result *_result, unsigned int idx )
{
	struct oracle_result *result = (struct oracle_result *) _result;
	i_info("gfn: name of field %d requested", idx);

	return (const char *)result->descriptor->S[idx];
}

static int driver_oracle_result_find_field(struct sql_result *_result, const char *field_name)
{
	struct oracle_result *result = (struct oracle_result *) _result;
	int i;
	char *field_conv = strdup( field_name );

	for (char *p = field_conv; (*p = (char)tolower(*p)); p++);

	i_info("ff: requested to find field [%s], (was [%s])", field_conv, field_name);
	for (i = 0; i < result->descriptor->F; i++) {
		if (strcmp(result->descriptor->S[i], field_conv) == 0)
			free( field_conv );
			return i;
	}
	free( field_conv );
	return -1;
}

static const unsigned char *driver_oracle_result_get_field_value_binary( struct sql_result *_result ATTR_UNUSED,
									unsigned int idx ATTR_UNUSED,
									size_t *size_r ATTR_UNUSED)
{
	/* FIXME */
	i_info("gvb: binary values are not supported yet");
	return NULL;
}

static const char *driver_oracle_result_get_field_value(struct sql_result *_result, unsigned int idx)
{
	struct oracle_result *result = (struct oracle_result *) _result;

	i_info("gvb: requested value of field %d", idx);
	return (const char *)result->descriptor->V[idx];
}

static const char *driver_oracle_result_find_field_value(struct sql_result *result, const char *field_name)
{
	int idx;

	i_info("ffv: requested value of field [%s]", field_name);
	idx = driver_oracle_result_find_field(result, field_name);
	if (idx < 0)
		return NULL;
	i_info("ffv: field [%s] has number %d. returning", field_name, idx);
	return driver_oracle_result_get_field_value(result, idx);
}

static const char *const *driver_oracle_result_get_values(struct sql_result *_result)
{
	struct oracle_result *result = (struct oracle_result *)_result;

	i_info("gv: requested all valued. returning SQLDA->V");
	return (const char *const *)result->descriptor->V;
}

static const char *driver_oracle_result_get_error(struct sql_result *_result)
{
	struct oracle_result *result = (struct oracle_result *)_result;

	i_info("ge: requested error.");

	return result->conn->error;
}

struct sql_db driver_oracle_db = {
	"oracle",

	MEMBER(v) {
		driver_oracle_init_v,
		driver_oracle_deinit_v,
		driver_oracle_get_flags,
		driver_oracle_connect_all,
		driver_oracle_escape_string,
		driver_oracle_exec,
		driver_oracle_query,
		driver_oracle_query_s,

		driver_oracle_transaction_begin,
		driver_oracle_transaction_commit,
		driver_oracle_transaction_commit_s,
		driver_oracle_transaction_rollback,

		driver_oracle_update
	}
};

struct sql_result driver_oracle_result = {
	MEMBER(v) {
		driver_oracle_result_free,
		driver_oracle_result_next_row,
		driver_oracle_result_get_fields_count,
		driver_oracle_result_get_field_name,
		driver_oracle_result_find_field,
		driver_oracle_result_get_field_value,
		driver_oracle_result_get_field_value_binary,
		driver_oracle_result_find_field_value,
		driver_oracle_result_get_values,
		driver_oracle_result_get_error
	}
};

static int
driver_oracle_result_error_next_row(struct sql_result *result ATTR_UNUSED)
{
	i_info("errnextrow: now rows available because of error");
        return -1;
}

struct sql_result driver_oracle_error_result = {
        MEMBER(v) {
                driver_oracle_result_free,
                driver_oracle_result_error_next_row,
                NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                driver_oracle_result_get_error
        }
};

void driver_oracle_init(void);
void driver_oracle_deinit(void);

void driver_oracle_init(void)
{
	sql_driver_register(&driver_oracle_db);
}

void driver_oracle_deinit(void)
{
	sql_driver_unregister(&driver_oracle_db);
}
#endif
